#include <stdio.h>
#include <stdlib.h>
#include <array.h>
#include <fido.h>
#include <logger.h>
#include <rule.h>
#include <util.h>
#include <version.h>
#ifdef HAVE_REGEX_H
# include <regex.h>
#else
# include <ereg.h>
#endif/*HAVE_REGEX_H*/
#include <joedog/defs.h>

#define MINLINES 10      
#define READSIZE BUFSIZ 
#define LINESIZE 128

struct FIDO_T
{
  char   *wfile;
  char   *rfile;
  char   *action;
  ARRAY  rules;
  LOG    logger;
  CONF   C;
};

struct pair {
  int  num;
  char *label;
  char *match;
};

private BOOLEAN __is_readable (FIDO this);
private long    __ticks (FIDO this, long offset);
private long    __get_offset (FIDO this);
private BOOLEAN __eregi (char *pattern, char *string);
private BOOLEAN __read_rules(FIDO this);

FIDO
new_fido(CONF C, const char *file)
{
  FIDO this;

  this = calloc(sizeof(*this),1);
  this->C      = C;
  this->wfile  = xstrdup(file);
  this->rfile  = xstrdup(conf_get_rules(this->C, this->wfile));
  this->action = xstrdup(conf_get_action(this->C, this->wfile));
  this->logger = new_logger(conf_get_log(this->C, this->wfile));
  this->rules  = new_array();
  if ((__read_rules(this)) == FALSE) {
    logger(this->logger, "%s [error] unable to read rules file: %s", program_name, this->rfile);
    if (is_verbose(C)) {
      NOTIFY(ERROR, "unable to read rules file: %s [exiting]", this->rfile);
    }
    exit(1);
  }
  return this;
}

FIDO
fido_destroy(FIDO this) {
  xfree(this->wfile);
  xfree(this->rfile);
  xfree(this->action);
  this->rules = array_destroy(this->rules);
  xfree(this);
  return NULL;
}

BOOLEAN 
start(FIDO this)
{
  int  rc = 0;
  long offset;

  if (__is_readable (this)) {
    offset = __get_offset(this);
    while (rc == 0 && offset >= 0) {
      offset = __ticks (this, offset);
      if (offset >= 0) {
        sleep (1);
      } else {
        logger(this->logger, "%s [error] set return to -7; offset value: %ld", program_name, offset);
        rc = -7;
      }
    }
  } else {
    logger(this->logger, "%s [error] unable to open \"%s\" for reading", program_name, this->wfile);
    if (is_verbose(this->C)) {
      NOTIFY(ERROR, "unable to open \"%s\" for reading", this->wfile);
    }
    rc = -4;
  }
  return rc == 0 ? FALSE : TRUE;
}

private BOOLEAN
__is_readable (FIDO this)
{
  FILE *fp;
  fp = fopen (this->wfile, "r");
  if (fp != NULL) {
    fclose (fp);
  }
  return (fp != NULL);
}

private long
__get_offset (FIDO this)
{
  FILE *fp;
  long offset;

  fp = fopen(this->wfile, "r");
  if (fp != NULL) {
    if (fseek (fp, 0, SEEK_END) == 0) {
      offset = ftell (fp);
      if (offset < 0) {
        offset = 0;
      }
    } else {
      offset = -1;
    }
    fclose (fp);
  } else {
    offset = -1;
  }
  return offset;
}

#define BLOCK 32 
char *fgetline(FILE *fp) 
{ 
  char *s; 
  char *tmp; 
  size_t count; 
  int ch; 
  
  for (count = 0, s = NULL; (ch = fgetc(fp)) != EOF && ch != '\n';count++) { 
    if (count%BLOCK == 0) { 
      tmp = realloc(s,count+BLOCK+1); 
      if (tmp == NULL) { 
        xfree(s); 
        return NULL; 
      } 
      s = tmp; 
    } 
    s[count] = ch; 
  } 
  if(s) s[count] = '\0'; 
  return s; 
}

//--ticks
private long
__ticks (FIDO this, long offset)
{
  FILE *fp;
  FILE *rp;
  int   ret;
  char *line;
  long  pos;
  char  out[BUFSIZ];


  fp = fopen (this->wfile, "r");
  if (fp != NULL) {
    pos = ftell(fp);
    fseek(fp, pos, SEEK_END);
    pos = ftell(fp);
    if (pos < offset) {
      /* file was truncated */
      offset = pos;
      VERBOSE(is_verbose(this->C), "%s was truncated; re-reading from the top", this->wfile);
    }
    if (fseek (fp, offset, SEEK_SET) == 0) {
      line = fgetline(fp);
      if (line != NULL) {
        int x;
        for (x = 0; x < array_length(this->rules); x++) {
          RULE r = (RULE)array_get(this->rules, x);
          if (__eregi(rule_get_pattern(r), line)) {
            if (is_daemon(this->C) == TRUE) {
              logger(this->logger,   "matched %s in %s", rule_get_pattern(r), this->wfile);
              VERBOSE(is_verbose(this->C), "matched %s in %s", rule_get_pattern(r), this->wfile);
            } else {
              VERBOSE(is_verbose(this->C), "matched: %s in %s", line, this->wfile);
            }
            /* XXX: WE NEED ERROR HANDLING 
                1. Does the executable exist?
                2. Does the executable run? 
                3. What is the result of the program?
             */
            char *cmd;
            if (rule_get_property(r) != NULL) {
              cmd = xmalloc(strlen(this->action)+strlen(rule_get_property(r))+5);
              memset(cmd, '\0', strlen(this->action)+strlen(rule_get_property(r))+5);
              snprintf(
                cmd, strlen(this->action)+strlen(rule_get_property(r))+5, "%s %s", this->action, rule_get_property(r)
              );
            } else {
              cmd = xstrdup(this->action);
            }
            rp = popen(cmd, "r");
            while (fgets(out, sizeof out, rp) != NULL) {
              chomp(out);
              if (conf_get_capture(this->C, this->wfile)) {
                logger(this->logger, "%s", out);
                VERBOSE(is_verbose(this->C), "%s", out);
              }
            }
            ret = pclose(rp);
            if (ret != 0) {
              logger(this->logger, "[error] %s exited with non-zero result: %d", this->action, ret);
              if (is_verbose(this->C)) {
                NOTIFY(ERROR, "%s exited with non-zero result: %d", this->action, ret);
              }
            }
            xfree(cmd);
          }
        }
        xfree(line);
      } else {
        offset = -1;
      }
      offset = ftell (fp);
    } else {
      logger(this->logger, "%s:%d: Can't seek to offset %ld.", __FILE__, __LINE__, offset);
      offset = -1;
    }
    fclose (fp);
  } else {
    logger(this->logger, "%s:%d: Can't re-open the file. Did it move?",__FILE__, __LINE__);
    offset = -1;
  }
  return offset;
}



static BOOLEAN
__eregi (char *match, char *string)
{
  int     res;
  regex_t reg;
  regmatch_t pm[10];
  char ebuf[128];
  const size_t nmatch = 10;

  if (!match)  return FALSE;
  if (!string) return FALSE;
  
  res = regcomp(&reg, match, REG_ICASE|REG_EXTENDED);
  if (res != 0) {
    regerror(res, &reg, ebuf, sizeof(ebuf));
    return FALSE;
  }
  res = regexec(&reg, string, nmatch, pm, 0);
  regfree(&reg);
  if (res == REG_NOMATCH){
    return FALSE;
  } else {
    return TRUE;
  }
  return FALSE;
}

BOOLEAN
__read_rules(FIDO this)
{
  int   num;
  char *line;
  FILE *fp;

  if ((fp = fopen(this->rfile, "r")) == NULL) {
     // the rules file is actually a rule
     RULE  R = new_rule(this->rfile);
     if (R) {
       array_push(this->rules, (RULE)R);
       return TRUE;
     } else {
       return FALSE;
     } 
  } else {
    while ((line = chomp_line(fp, &line, &num)) != NULL) {
      RULE  R;
      char *lineptr;
      char *lineval;
      lineptr = lineval = xstrdup(line);
      while (*lineptr && !ISCOMMENT(*lineptr))
        lineptr++;
      *lineptr++=0;
      R = new_rule(lineval);
      if (R) {
        array_push(this->rules, (RULE)R);
      }
      xfree(lineval);
    }
    return TRUE;
  }
  return FALSE;
}


